"""
This module provides functionality for interacting with CAMV.

Currently limited to importing and outputing scan lists.
"""

# Built-ins
from __future__ import absolute_import, division

from collections import OrderedDict
from functools import partial
import logging
import multiprocessing
import os
import tempfile
import shutil

from . import compare, fragments, gen_sequences, ms_labels
from pycamv.search import search
from pycamv.export import export
from pycamv.matlab import camv_mat
from pycamv.scan import scans
from pycamv.scan import scan_list as sl
from pycamv.utils import LenGen


LOGGER = logging.getLogger("pycamv.validate")


def _remap_pst(pep_mods):
    return [
        (
            count,
            mod,
            letters + (
                ("Y",)
                if (mod == "Phospho" and set(letters) == set(["S", "T"])) else
                ()
            ),
        )
        for count, mod, letters in pep_mods
    ]


def _map_seq(kv, limit_comb=False):
    lowpriority()

    pep_query = kv

    gen = gen_sequences.gen_possible_seq(
        pep_query.pep_seq,
        pep_query.pep_mods,
    )

    if limit_comb:
        return (
            pep_query,
            tuple(
                seq
                for index, seq in zip(
                    range(gen_sequences.MAX_NUM_COMB),
                    gen,
                )
            )
        )
    else:
        return (
            pep_query,
            tuple(gen),
        )


def _to_str(seq):
    return "".join(
        letter.lower()
        if [i for i in mods if i not in ms_labels.LABEL_NAMES]
        else letter.upper()
        for letter, mods in seq[1:-1]
    )


def _close_scans(ms_datas, out_dir=None):
    LOGGER.info("Removing directory of temporary files.")

    for ms_data in ms_datas:
        for raw in ms_data.values():
            raw.info['fileObject'].close()
            raw.seeker.close()

    if out_dir:
        shutil.rmtree(out_dir)


def _get_window_coverage(pep_query, scan_query, precursor_win):
    precursor_win = [
        mz
        for mz, _ in precursor_win
        if mz >= scan_query.isolation_mz - scan_query.window_offset[0] and
        mz <= scan_query.isolation_mz + scan_query.window_offset[1]
    ]
    return (
        max([
            c13
            for c13 in range(
                1, 1 + round(scan_query.window_offset[1] * pep_query.pep_exp_z)
            )
            if min([
                1e6 * abs(
                    scan_query.isolation_mz +
                    fragments.DELTA_C13 * c13 / pep_query.pep_exp_z -
                    mz
                ) / mz
                for mz in precursor_win
            ] + [compare.MS_TOL]) < compare.MS_TOL
        ] + [0])
        if scan_query.window_offset
        else 0
    )


def _map_frag_compare(kv):
    lowpriority()

    try:
        (
            pep_query, scan_query, sequence,
            ms_two_data, ms_data,
            validation_data, auto_maybe,
        ) = kv

        ms_scan = ms_data[scan_query.basename][scan_query.precursor_scan]

        if ms_scan["id"] != scan_query.precursor_scan:
            LOGGER.warning(
                "{} #{}: Precursor scan id different from expected: {} != {}"
                .format(
                    pep_query.query,
                    pep_query.scan,
                    ms_scan["id"],
                    scan_query.precursor_scan,
                )
            )

        precursor_win = scans.get_precursor_peak_window(
            scan_query, ms_scan
        )

        # Get the max C13 peak found within the isolation window
        window_coverage = _get_window_coverage(
            pep_query, scan_query, precursor_win,
        )

        frag_ions = fragments.fragment_ions(
            sequence, pep_query.pep_exp_z,
            c13_num=scan_query.c13_num + window_coverage,
        )

        # Compare MS^2 data with predicted fragment ions
        ms_two_scan = ms_two_data[pep_query.basename][pep_query.scan]

        if ms_two_scan["id"] != scan_query.scan:
            LOGGER.warning(
                "{} #{}: MS^2 scan id different from expected: {} != {}"
                .format(
                    pep_query.query,
                    pep_query.scan,
                    ms_two_scan["id"],
                    scan_query.scan,
                )
            )

        peaks = compare.compare_spectra(
            ms_two_scan, frag_ions,
            tol=compare.COLLISION_TOLS[scan_query.collision_type],
        )

        # XXX: Unlabeled runs?
        quant_scan = (
            ms_two_data[pep_query.basename][pep_query.quant_scan]
            if pep_query.scan != pep_query.quant_scan else
            ms_two_scan
        ) if pep_query.quant_scan is not None else None

        label_win = scans.get_label_peak_window(pep_query, quant_scan)

        choice = validation_data.get((pep_query.scan, _to_str(sequence)), None)

        if not choice and auto_maybe:
            if (
                pep_query.rank_pos is not None and
                pep_query.rank_pos.get(1, None) == set(
                    (pos, mod)
                    for pos, (_, mods) in enumerate(sequence[1:-1])
                    for mod in mods
                )
            ):
                choice = "maybe"

        return (
            pep_query, tuple(sequence), choice,
            peaks, precursor_win, label_win,
        )
    except:
        _close_scans([ms_data, ms_two_data])
        raise


def fill_map_frag_compare(
    sequence_mapping, scan_mapping,
    ms_two_data, ms_data,
    queue, cpu_count,
    validation_data, auto_maybe,
):
    pool = multiprocessing.Pool(
        processes=cpu_count - 1,
    )

    total_num_seq = sum(len(i) for i in sequence_mapping.values())

    try:
        peak_hits = pool.imap_unordered(
            func=_map_frag_compare,
            iterable=LenGen(
                gen=(
                    (
                        pep_query,
                        scan_mapping[pep_query],
                        sequence,
                        ms_two_data,
                        ms_data,
                        validation_data,
                        auto_maybe,
                    )
                    for pep_query, sequences in sequence_mapping.items()
                    for sequence in sequences
                ),
                len=total_num_seq,
            ),
        )

        for item in peak_hits:
            queue.put(item)

        pool.close()
    except:
        pool.terminate()
        raise
    finally:
        pool.join()
        _close_scans([ms_data, ms_two_data])


# Taken from https://stackoverflow.com/questions/1023038/
def lowpriority():
    """ Set the priority of the process to below-normal."""

    import sys
    try:
        sys.getwindowsversion()
    except AttributeError:
        isWindows = False
    else:
        isWindows = True

    if isWindows:
        # Based on:
        #   "Recipe 496767: Set Process Priority In Windows" on ActiveState
        #   http://code.activestate.com/recipes/496767/
        import win32api
        import win32process
        import win32con

        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(
            handle, win32process.BELOW_NORMAL_PRIORITY_CLASS,
        )
    else:
        import os

        os.nice(1)


def validate_spectra(
    search_path,
    raw_paths=None,
    scans_path=None,
    scan_list=None,
    mat_sessions=None,
    out_path=None,
    cpu_count=None,
    reprocess=False,
    auto_maybe=False,
):
    """
    Generate CAMV web page for validating spectra.

    Parameters
    ----------
    search_path : str
    raw_paths : list of str
    scans_path : str, optional
    scan_list : list of int, optional
    mat_sessions : list of str, optional
    out_path : str, optional
    cpu_count : int, optional
    reprocess : bool, optional
    auto_maybe : bool, optional
    """
    if cpu_count is None:
        try:
            cpu_count = multiprocessing.cpu_count()
        except NotImplementedError:
            cpu_count = 2

    LOGGER.debug("Validating data set with {} cpus".format(cpu_count))

    if raw_paths is None:
        raw_paths = []

    if scan_list is None:
        scan_list = []

    if scans_path is not None:
        scan_list += sl.load_scan_list(scans_path)

    if out_path is None:
        out_path = os.path.splitext(search_path)[0] + ".camv.gz"

    # Read peptide search file
    fixed_mods, var_mods, pep_queries = search.read_search_file(search_path)

    LOGGER.info("Found info for {} peptide queries".format(len(pep_queries)))

    # Optionally filter queries using a scan list
    if scan_list:
        pep_queries = [
            query
            for query in pep_queries
            if query.scan in scan_list
        ]

    # Remap pST -> pSTY
    LOGGER.info("Remapping pST -> pSTY")
    for pep_query in pep_queries:
        pep_query.pep_var_mods = _remap_pst(pep_query.pep_var_mods)
        pep_query.pep_fixed_mods = _remap_pst(pep_query.pep_fixed_mods)

    # Get scan data from RAW file
    required_raws = set(query.basename for query in pep_queries)
    base_raw_paths = [os.path.basename(path) for path in raw_paths]
    missing = []

    for base_raw in required_raws:
        if base_raw in base_raw_paths:
            continue

        for base_dir in [
            os.path.dirname(search_path),
            os.path.join(os.path.dirname(search_path), "..", "MS RAW")
        ]:
            local_raw_path = os.path.join(base_dir, base_raw)
            if os.path.exists(local_raw_path):
                raw_paths.append(local_raw_path)
                break
        else:
            missing.append(base_raw)

    if missing:
        raise Exception(
            "Unable to find {} in input RAW files: {}"
            .format(missing, base_raw_paths)
        )

    # Generate sequences
    LOGGER.info(
        "Generating all possible sequence-modification combinations."
    )
    pool = multiprocessing.Pool(
        processes=cpu_count,
    )
    try:
        sequence_mapping = dict(
            pool.imap_unordered(
                partial(_map_seq, limit_comb=not reprocess),
                pep_queries,
            )
        )
        pool.close()
    except:
        pool.terminate()
        raise
    finally:
        pool.join()

    total_num_seq = sum(len(i) for i in sequence_mapping.values())

    LOGGER.info(
        "Generating fragment ions for {} possible sequences."
        .format(
            total_num_seq,
        )
    )

    # Import validation data from CAMV-Matlab
    validation_data = {}

    if mat_sessions:
        LOGGER.info("Loading validation data from {}".format(mat_sessions))

        validation_data = camv_mat.load_mat_validation(mat_sessions)

        LOGGER.debug(
            "Loaded {} sequence validations from MATLAB sessions"
            .format(len(validation_data))
        )

    # Import raw scan data
    out_dir = tempfile.mkdtemp()

    LOGGER.info("Getting scan data.")

    scan_queries, ms_two_data, ms_data = scans.get_scan_data(
        raw_paths, pep_queries, out_dir,
    )

    try:
        LOGGER.info("Found data for {} scans".format(len(scan_queries)))

        scan_mapping = OrderedDict(
            (pep_query, scan_query)
            for pep_query, scan_query in zip(pep_queries, scan_queries)
        )

        # Generate fragments and assign peaks to fragments
        LOGGER.info(
            (
                "Comparing predicted to actual peaks for {} spectra "
                " ({} pep-scan combinations)."
            ).format(
                len(scan_mapping),
                total_num_seq,
            )
        )

        queue = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=fill_map_frag_compare,
            args=(
                sequence_mapping,
                scan_mapping,
                ms_two_data,
                ms_data,
                queue,
                cpu_count,
                validation_data,
                auto_maybe,
            ),
        )
        process.start()

        # XXX: Determine SILAC precursor masses?

        # XXX: Remove precursor contaminated scans from validation list?

        # Check each assignment to each scan

        # Output data
        export.export_to_sql(
            os.path.splitext(out_path)[0] + ".db",
            queue, scan_mapping,
            search_path, raw_paths,
            total_num_seq=total_num_seq,
            reprocess=reprocess,
        )
    except:
        process.terminate()
        raise
    finally:
        process.join()
        _close_scans([ms_data, ms_two_data], out_dir=out_dir)

    LOGGER.info(
        "Exported {} total peptide-scan combinations"
        .format(total_num_seq)
    )



# -*- coding: utf-8 -*-

import argparse
import logging
import multiprocessing
import os
import sys

from . import fragment, scan, search, __version__


LOGGER = logging.getLogger("pycamv.main")

RAW_EXTS = [".raw", ".mgf", ".d", ".wiff"]
SEARCH_EXTS = list(search.search.BACKENDS.keys())
SCANS_EXTS = list(scan.scan_list.BACKENDS.keys())


def _parse_args(args):
    """
    Parses arguments from a argv format.

    Parameters
    ----------
    args : list of str

    Returns
    -------
    argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        prog="PyCAMVerter",
        description="Aww yeah, mass specs!",
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        help="Increase verbosity of output.",
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        help="Decrease verbosity of output.",
    )
    parser.add_argument(
        '-V', '--version',
        action="version",
        version="%(prog)s {}".format(__version__),
    )
    parser.add_argument(
        "--cpus",
        type=int,
        help="Limit the number of concurrent processes."
    )
    parser.add_argument(
        "--reprocess",
        action="store_true",
        help="Reprocess a set of scans, without limiting ptm combinations.",
    )
    parser.add_argument(
        "--no-auto-maybe",
        action="store_false",
        dest="auto_maybe",
        help="Don't auto-assign all peptides with best MASCOT rank as 'maybe'",
    )
    parser.add_argument(
        "--raw-paths",
        nargs="+",
        help="Raw data file(s) containing mass spec data.",
    )
    parser.add_argument(
        "--search-path",
        help="MASCOT or ProteomeDiscoverer search files."
    )
    parser.add_argument(
        "--scans-path",
        help=".xlsx or .csv file listing scans to select for validation.",
    )
    parser.add_argument(
        "--scans",
        nargs="*", type=int,
        help="Individual scans to select for validation.",
    )
    parser.add_argument(
        "--mat-sessions",
        nargs="+",
        help="Path to CAMV-Matlab session files.",
    )
    parser.add_argument(
        "--out-path",
        help="Output path for CAMV export.",
    )
    parser.add_argument(
        'files',
        nargs='*',
        help="Raw, search, or scan list files, determined by file extension."
    )
    return parser, parser.parse_args(args)


def main(args):
    parser, args = _parse_args(args)

    verbosity = (args.verbose or 0) - (args.quiet or 0)

    if verbosity <= -2:
        level = logging.CRITICAL
    elif verbosity == -2:
        level = logging.ERROR
    elif verbosity == -1:
        level = logging.WARNING
    elif verbosity == 0:
        level = logging.INFO
    elif verbosity > 0:
        level = logging.DEBUG

    logger = logging.getLogger('pycamv')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    fh = logging.FileHandler('pycamv.log')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    LOGGER.debug(sys.argv)
    LOGGER.debug(args)

    if args.search_path is None:
        searches = [
            i for i in args.files
            if os.path.splitext(i.lower())[1] in SEARCH_EXTS
        ]

        if len(searches) == 1:
            args.search_path = searches[0]

    if args.raw_paths is None:
        raws = [
            i for i in args.files
            if os.path.splitext(i.lower())[1] in RAW_EXTS
        ]

        if len(raws) > 0:
            if args.raw_paths:
                args.raw_paths += raws
            else:
                args.raw_paths = raws

    if args.scans_path is None:
        scans = [
            i for i in args.files
            if os.path.splitext(i.lower())[1] in SCANS_EXTS
        ]

        if len(scans) == 1:
            args.scans_path = scans[0]

    if (
        args.search_path is None
    ):
        parser.print_help()
        raise Exception(
            "Missing search input path"
        )

    fragment.validate.validate_spectra(
        search_path=args.search_path,
        raw_paths=args.raw_paths,
        scans_path=args.scans_path,
        scan_list=args.scans,
        mat_sessions=args.mat_sessions,
        out_path=args.out_path,
        reprocess=args.reprocess,
        auto_maybe=args.auto_maybe,
        cpu_count=args.cpus,
    )


if __name__ == "__main__":
    multiprocessing.freeze_support()
    # is_frozen_executable = getattr(sys, u'frozen', False)
    #
    # if is_frozen_executable:
    #     os.chdir(os.path.dirname(sys.executable))
    # else:
    #     os.chdir(os.path.join(os.path.dirname(__file__), u'..'))

    try:
        main(sys.argv[1:])
    except Exception as e:
        LOGGER.error("PyCAMV Converter has crashed!", exc_info=True)
        raise e




import re

RE_DISCOVERER_DESCRIPTION = re.compile(
    r"^>sp\|([\dA-Za-z]+)\|([\dA-Za-z_]+) (.+) OS=(.+)$"
)
RE_MASCOT_DESCRIPTION = re.compile(
    r"^(.+) OS=(.+)$"
)
RE_DYN_MODS = re.compile(r"((\d+) )?(.+) \((.+)\)")
RE_BY_ION_POS = re.compile(r"([abcxyz])_\{(\d+)\}")
RE_B_Y_IONS = re.compile(r"([abcxyz]_\{[0-9]+\})(.*)\^\{\+\}")
RE_SCAN_NUM = re.compile(r"(scans:|Cmpd_)(\d+)")
RE_COLLISION_TYPE = re.compile(r".*@([A-Za-z]+)\d+")
RE_PRECURSOR_SCAN = re.compile(r"scan=(\d+)")



"""
This module provides functionality for comparing spectra with a list of
fragments.
"""

from __future__ import absolute_import, division


MS_TOL = 10
CID_TOL = 1000
HCD_TOL = 10

COLLISION_TOLS = {
    "CID": CID_TOL,
    "HCD": HCD_TOL,
}


class PeptideHit:
    """
    Attributes
    ----------
    mz : float
    intensity : float
    name : str or None
    score : int or None
    predicted_mz : float or None
    match_list : dict of str, tuple of (float, float)
    num_losses : int
    """
    def __init__(
        self, mz, intensity,
        name=None, score=None, predicted_mz=None, match_list=None,
    ):
        self.mz = mz
        self.intensity = intensity
        self.name = name
        self.score = score
        self.predicted_mz = predicted_mz
        self.match_list = match_list
        self.num_losses = name.count("-") if name else 0

    @property
    def frac_error(self):
        if self.predicted_mz is None:
            return 0

        return abs(self.predicted_mz - self.mz) / self.predicted_mz


def _calculate_ion_score(ion_name):
    score = 0

    if ion_name.startswith("MH"):
        score += 12
    elif ion_name.startswith("b_") or ion_name.startswith("y_"):
        score += 10

    score -= ion_name.count("-")

    return score


def compare_spectra(
    spectra, frag_ions,
    tol=None,
):
    """
    Parameters
    ----------
    spectra : :class:`pymzml.spec.Spectrum<spec.Spectrum>`
    frag_ions : dict of str, float
    tol : float, optional

    Returns
    -------
    list of :class:`PeptideHit<pycamv.compare.PeptideHit>`
        Peak assignments for each peak in spectra.
    """
    if tol is None:
        tol = CID_TOL

    max_y = max(spectra.i)

    # Iterate over each peak, looking for fragments to assign to it
    out = []

    # Sort frag ions, so we only calculate ppms on peaks local to that area
    # of the spectrum
    frag_ions = list(sorted(frag_ions.items(), key=lambda x: x[1]))
    last_frag_index = 0
    last_mz = 0

    # Reprofiled Peaks? Centroided Peaks?
    for mz, intensity in spectra.centroidedPeaks:
        assert mz >= last_mz
        last_mz = mz

        peak_candidates = {}
        min_index_set = False

        for index, (ion_name, ion_mz) in enumerate(
            frag_ions[last_frag_index:]
        ):
            # non-abs ppm should be monoisotopically increasing around mz
            if 1e6 * abs(ion_mz - mz) / ion_mz < 1.5 * tol:
                peak_candidates[ion_name] = ion_mz, abs(ion_mz - mz)

                if not min_index_set:
                    last_frag_index = index
                    min_index_set = True
            elif 1e6 * (ion_mz - mz) / ion_mz > 1.5 * tol:
                break

        # peak_candidates = {
        #     ion_name: (ion_mz, abs(ion_mz - mz))
        #     for ion_name, ion_mz in frag_ions
        #     if 1e6 * abs(ion_mz - mz) / ion_mz < 1.5 * tol
        # }

        if not peak_candidates:
            # XXX: AA Identification?
            out.append(PeptideHit(mz, intensity))
            continue

        # Calculate ion scores based on their names
        scores = {
            ion_name: _calculate_ion_score(ion_name)
            for ion_name in peak_candidates
        }

        # XXX: 0.5? 1
        closest_ion = min(peak_candidates, key=lambda x: peak_candidates[x][1])
        scores[closest_ion] += 1

        best_ion = max(scores, key=lambda x: scores[x])

        retain = (
            intensity > 0.1 * max_y or
            best_ion.startswith("MH") or
            best_ion.startswith("b_") or
            best_ion.startswith("y_") or
            (
                best_ion.count("-") == 0 and
                intensity > 0.025 * max_y
            )
        )

        if not retain:
            best_ion = None

        out.append(
            PeptideHit(
                mz,
                intensity,
                name=best_ion,
                score=scores.get(best_ion, None),
                predicted_mz=peak_candidates.get(best_ion, None),
                match_list=peak_candidates,
            )
        )

    return out



"""
Main module for running pycamv from the commandline.
"""

from __future__ import absolute_import, division

import multiprocessing
import sys

from . import main

if __name__ == "__main__":
    multiprocessing.freeze_support()

    main.main(sys.argv[1:])



import os
import sys

# Module multiprocessing is organized differently in Python 3.4+
try:
    # Python 3.4+
    if sys.platform.startswith('win'):
        import multiprocessing.popen_spawn_win32 as forking
    else:
        import multiprocessing.popen_fork as forking
except ImportError:
    import multiprocessing.forking as forking

if sys.platform.startswith('win'):
    # First define a modified version of Popen.
    class _Popen(forking.Popen):
        def __init__(self, *args, **kw):
            if hasattr(sys, 'frozen'):
                # We have to set original _MEIPASS2 value from sys._MEIPASS
                # to get --onefile mode working.
                os.putenv('_MEIPASS2', sys._MEIPASS)
            try:
                super(_Popen, self).__init__(*args, **kw)
            finally:
                if hasattr(sys, 'frozen'):
                    # On some platforms (e.g. AIX) 'os.unsetenv()' is not
                    # available. In those cases we cannot delete the variable
                    # but only set it to the empty string. The bootloader
                    # can handle this case.
                    if hasattr(os, 'unsetenv'):
                        os.unsetenv('_MEIPASS2')
                    else:
                        os.putenv('_MEIPASS2', '')

    # Second override 'Popen' class with our modified version.
    forking.Popen = _Popen




import logging
import sqlite3

from . import migrations
from pycamv import regexes, utils, version
from pycamv.fragment import gen_sequences, ms_labels

LOGGER = logging.getLogger("pycamv.sql")

DB_EXTS = [".db", ".sql"]
DATA_VERSION = "1.1.0"

CAMV_SCHEMA = """
PRAGMA foreign_keys = ON;
PRAGMA journal_mode = OFF;
PRAGMA synchronous = OFF;
PRAGMA temp_store = MEMORY;

-- Individual protein names (i.e. Src)
CREATE TABLE IF NOT EXISTS proteins
(
    protein_id              integer primary key autoincrement not null,
    protein_name            text,
    protein_accession       text,
    protein_uniprot         text,
    full_sequence           text,
    UNIQUE(protein_accession)
);

-- Pre-processed set of protein names (i.e. Cdk1 / Cdk2 / Cdk3)
CREATE TABLE IF NOT EXISTS protein_sets
(
    protein_set_id          integer primary key autoincrement not null,
    protein_set_name        text,
    protein_set_accession   text,
    protein_set_uniprot     text,
    UNIQUE(protein_set_accession)
);

-- Individual peptide sequences (i.e. IVLEYK)
CREATE TABLE IF NOT EXISTS peptides
(
    peptide_id              integer primary key autoincrement not null,
    protein_set_id          integer not null,
    protein_set_offsets     text,
    peptide_seq             text,
    FOREIGN KEY(protein_set_id) REFERENCES protein_sets(protein_set_id)
    UNIQUE(peptide_seq)
);

-- Mapping between proteins and peptides
CREATE TABLE IF NOT EXISTS protein_peptide
(
    prot_pep_id             integer primary key autoincrement not null,
    peptide_id              integer not null,
    protein_id              integer not null,
    peptide_offset          integer,
    FOREIGN KEY(peptide_id) REFERENCES peptides(peptide_id),
    FOREIGN KEY(protein_id) REFERENCES proteins(protein_id),
    UNIQUE(peptide_id, protein_id)
);

-- Peptides with unmapped modifications (i.e. +1 pY)
CREATE TABLE IF NOT EXISTS mod_states
(
    mod_state_id            integer primary key autoincrement not null,
    peptide_id              integer not null,
    mod_desc                text,
    num_comb                integer,
    FOREIGN KEY(peptide_id) REFERENCES peptides(peptide_id),
    UNIQUE(peptide_id, mod_desc)
);

-- Peptides with modifications exactly positioned (i.e pY114)
CREATE TABLE IF NOT EXISTS ptms
(
    ptm_id                  integer primary key autoincrement not null,
    mod_state_id            integer not null,
    name                    text,
    full_name               text,
    FOREIGN KEY(mod_state_id) REFERENCES mod_states(mod_state_id),
    UNIQUE(mod_state_id, full_name)
);

-- Raw files sourced for each scan
CREATE TABLE IF NOT EXISTS files
(
    file_id                 integer primary key autoincrement not null,
    filename                text,
    UNIQUE(filename)
);

-- Blobs of scan data, for full ms2, quantification window, and precursor data
CREATE TABLE IF NOT EXISTS scan_data
(
    data_id                 integer primary key autoincrement not null,
    scan_id                 integer not null,
    data_type               text,
    data_blob               blob,
    FOREIGN KEY(scan_id) REFERENCES scans(scan_id),
    UNIQUE(scan_id, data_type)
);

-- Set of peaks that are used for quantitation
CREATE TABLE IF NOT EXISTS quant_mz
(
    quant_mz_id             integer primary key autoincrement not null,
    label_name              text,
    UNIQUE(label_name)
);

-- Individual peak / mz names used for quantitation
CREATE TABLE IF NOT EXISTS quant_mz_peaks
(
    quant_mz_peak_id        integer primary key autoincrement not null,
    quant_mz_id             integer not null,
    mz                      real not null,
    peak_name               text not null,
    FOREIGN KEY(quant_mz_id) REFERENCES quant_mz(quant_mz_id),
    UNIQUE(quant_mz_id, mz, peak_name)
);

CREATE TABLE IF NOT EXISTS scans
(
    scan_id                 integer primary key autoincrement not null,
    scan_num                integer not null,
    charge                  integer not null,
    pep_exp_mz              integer not null,
    collision_type          text not null,
    precursor_mz            real not null,
    isolation_window_lower  real,
    isolation_window_upper  real,
    quant_mz_id             integer,
    c13_num                 integer,
    file_id                 integer,
    truncated               integer,
    FOREIGN KEY(quant_mz_id) REFERENCES quant_mz(quant_mz_id),
    FOREIGN KEY(file_id) REFERENCES files(file_id),
    UNIQUE(scan_num, file_id)
);

CREATE TABLE IF NOT EXISTS scan_ptms
(
    scan_ptm_id             integer primary key autoincrement not null,
    scan_id                 integer not null,
    ptm_id                  integer not null,
    choice                  text,
    mascot_score            real,
    FOREIGN KEY(scan_id) REFERENCES scans(scan_id),
    FOREIGN KEY(ptm_id) REFERENCES ptms(ptm_id),
    UNIQUE(scan_id, ptm_id)
);

CREATE TABLE IF NOT EXISTS fragments
(
    fragment_id             integer primary key autoincrement not null,
    peak_id                 integer not null,
    scan_ptm_id             integer not null,
    name                    text not null,
    display_name            text not null,
    mz                      real not null,
    intensity               real not null,
    best                    integer,
    ion_type                text,
    ion_pos                 integer,
    FOREIGN KEY(scan_ptm_id) REFERENCES scan_ptms(scan_ptm_id)
    UNIQUE(peak_id, scan_ptm_id, name)
);
CREATE INDEX IF NOT EXISTS fragments_idx ON fragments(scan_ptm_id);
CREATE INDEX IF NOT EXISTS fragments_idx ON fragments(peak_id, scan_ptm_id);

CREATE TABLE IF NOT EXISTS camv_meta
(
    key                     text not null,
    val                     text,
    UNIQUE(key, val)
);
"""


def create_tables(cursor):
    cursor.executescript(CAMV_SCHEMA)
    insert_camv_meta(cursor)
    cursor.connection.commit()


def run_migrations(cursor):
    rows = cursor.execute(
        """
        SELECT camv_meta.key, camv_meta.val
        FROM camv_meta
        WHERE camv_meta.key IN (?)
        """,
        [
            "camvDataVersion",
        ],
    )

    camv_data_version = None

    for key, val in rows:
        if key == "camvDataVersion":
            camv_data_version = val
        else:
            raise Exception(
                "Unexpected key in camv_meta: {} - {}".format(key, val)
            )

    if camv_data_version is None:
        raise Exception(
            "Unable to determine CAMV data version of existing database",
        )

    if camv_data_version != DATA_VERSION:
        migrations.run_migrations(camv_data_version, DATA_VERSION, cursor)


def _insert_or_update_row(
    cursor, table, id, data,
    unique_on=None, update=False,
):
    if unique_on is None:
        unique_on = list(data.keys())

    row_id = None

    for row in cursor.execute(
        """
        SELECT ({})
        FROM {}
        WHERE {}
        """.format(
            id,
            table,
            " and ".join("{}=(?)".format(i) for i in unique_on)
        ),
        [data[i] for i in unique_on],
    ):
        row_id = row[0]

    if row_id is None:
        cursor.execute(
            """
            INSERT OR IGNORE INTO {}
            ({})
            VALUES ({})
            """.format(
                table,
                ", ".join(data.keys()),
                ",".join("?" for i in data.keys()),
            ),
            list(data.values()),
        )
        row_id = cursor.lastrowid
    elif update:
        keys = [i for i in data.keys() if i not in unique_on]
        cursor.execute(
            """
            UPDATE {}
            SET {}
            WHERE {}
            """.format(
                table,
                ", ".join("{}=?".format(i) for i in keys),
                " and ".join("{}=?".format(i) for i in unique_on),
            ),
            [data[i] for i in keys] + [data[i] for i in unique_on],
        )

    return row_id


def insert_protein(cursor, query):
    return [
        _insert_or_update_row(
            cursor, "proteins", "protein_id",
            {
                "protein_name": prot_desc,
                "protein_accession": access,
                "protein_uniprot": uniprot,
                "full_sequence": seq,
            },
            unique_on=["protein_accession"],
            update=True,
        )
        for prot_desc, access, uniprot, seq in zip(
            query.prot_descs, query.accessions,
            query.uniprot_accessions, query.full_seqs,
        )
    ]


def insert_protein_set(cursor, query):
    return _insert_or_update_row(
        cursor, "protein_sets", "protein_set_id",
        {
            "protein_set_name": " / ".join(
                i[0]
                for i in sorted(
                    zip(query.prot_descs, query.accessions),
                    key=lambda x: x[1]
                )
            ),
            "protein_set_accession": " / ".join(
                sorted(query.accessions)
            ),
            "protein_set_uniprot": ";".join(
                i[0]
                for i in sorted(
                    zip(query.uniprot_accessions, query.accessions),
                    key=lambda x: x[1],
                )
            ),
        },
        unique_on=["protein_set_accession"],
        update=True,
    )


def insert_peptide(cursor, query, prot_set_id):
    return _insert_or_update_row(
        cursor, "peptides", "peptide_id",
        {
            "peptide_seq": query.pep_seq,
            "protein_set_id": prot_set_id,
            "protein_set_offsets": ";".join(
                str(i[0])
                for i in sorted(
                    zip(query.pep_offsets, query.accessions),
                    key=lambda x: x[1],
                )
            ),
        },
        unique_on=["peptide_seq", "protein_set_id"],
        update=True,
    )


def insert_pep_prot(cursor, pep_id, prot_ids, prot_offsets):
    return [
        _insert_or_update_row(
            cursor, "protein_peptide", "prot_pep_id",
            {
                "peptide_id": pep_id,
                "protein_id": prot_id,
                "peptide_offset": offset,
            },
            unique_on=[
                "peptide_id",
                "protein_id",
            ],
            update=True,
        )
        for prot_id, offset in zip(prot_ids, prot_offsets)
    ]


def _get_labels_mz(query):
    def _within(val, bounds):
        return val >= bounds[0] and val <= bounds[1]
    return [
        (mz, name)
        for mod in set(query.get_label_mods)
        for mz, name in zip(
            ms_labels.LABEL_MASSES.get(mod, []),
            ms_labels.LABEL_NAMES.get(mod, []),
        )
        if _within(mz, ms_labels.LABEL_MZ_WINDOW.get(mod, [mz, mz]))
    ]


def insert_quant_mz(cursor, query):
    quant_mz_id = _insert_or_update_row(
        cursor, "quant_mz", "quant_mz_id",
        {
            "label_name": ";".join(sorted(set(query.get_label_mods))),
        },
    )

    cursor.executemany(
        """
        INSERT OR IGNORE INTO quant_mz_peaks
        (
            quant_mz_id,
            mz,
            peak_name
        )
        VALUES (?, ?, ?)
        """,
        [
            (
                quant_mz_id,
                mz,
                name
            )
            for mz, name in _get_labels_mz(query)
        ],
    )

    return quant_mz_id


def insert_file(cursor, query):
    return _insert_or_update_row(
        cursor, "files", "file_id",
        {
            "filename": query.filename,
        },
    )


def _peaks_to_blob(peaks):
    return sqlite3.Binary(
        ";".join(
            "{},{}".format(
                i.mz if hasattr(i, "mz") else i[0],
                i.intensity if hasattr(i, "intensity") else i[1],
            )
            for i in peaks
        ).encode("utf-8")
    )


def insert_peaks(cursor, peaks, scan_id):
    return _insert_or_update_row(
        cursor, "scan_data", "data_id",
        {
            "scan_id": scan_id,
            "data_type": "ms2",
            "data_blob": _peaks_to_blob(peaks),
        },
        unique_on=["scan_id", "data_type"],
    )


def insert_precursor_peaks(cursor, scan_query, precursor_win, scan_id):
    return _insert_or_update_row(
        cursor, "scan_data", "data_id",
        {
            "scan_id": scan_id,
            "data_type": "precursor",
            "data_blob": _peaks_to_blob(precursor_win),
        },
        unique_on=["scan_id", "data_type"],
    )


def insert_quant_peaks(cursor, query, label_win, scan_id):
    return _insert_or_update_row(
        cursor, "scan_data", "data_id",
        {
            "scan_id": scan_id,
            "data_type": "quant",
            "data_blob": _peaks_to_blob(label_win),
        },
        unique_on=["scan_id", "data_type"],
    )


def insert_scans(
    cursor, query, scan_query,
    quant_mz_id, file_id,
    reprocessed=False, c13_num=None,
):
    return _insert_or_update_row(
        cursor, "scans", "scan_id",
        {
            "scan_num": query.scan,
            "charge": query.pep_exp_z,
            "pep_exp_mz": query.pep_exp_mz,
            "collision_type": scan_query.collision_type,
            "precursor_mz": scan_query.isolation_mz,
            "isolation_window_lower": scan_query.window_offset[0],
            "isolation_window_upper": scan_query.window_offset[1],
            "c13_num": c13_num or scan_query.c13_num,
            "quant_mz_id": quant_mz_id,
            "file_id": file_id,
            "truncated": (
                not reprocessed and query.num_comb > gen_sequences.MAX_NUM_COMB
            ),
        },
        unique_on=["scan_num", "file_id"],
        update=reprocessed,
    )


def insert_scan_ptms(cursor, query, scan_id, ptm_id, choice=None):
    return _insert_or_update_row(
        cursor, "scan_ptms", "scan_ptm_id",
        {
            "scan_id": scan_id,
            "ptm_id": ptm_id,
            "mascot_score": query.pep_score,
            "choice": choice,
        },
        unique_on=["scan_id", "ptm_id"],
    )


def _get_mods_description(pep_seq, mod_state):
    return ("+ " if mod_state else "") + " - ".join(
        "{} {}{}".format(count, mod[0].lower(), "".join(letters))
        for count, mod, letters in mod_state
    )


def insert_mod_state(cursor, query, peptide_id):
    return _insert_or_update_row(
        cursor, "mod_states", "mod_state_id",
        {
            "peptide_id": peptide_id,
            "mod_desc": _get_mods_description(
                query.pep_seq,
                tuple(query.pep_var_mods),
            ),
            "num_comb": query.num_comb,
        },
        unique_on=[
            "peptide_id",
            "mod_desc",
        ]
    )


def _pep_mod_name(pep_seq, mods):
    return "".join(
        letter.lower() if mods else letter.upper()
        for letter, mods in zip(pep_seq, mods[1:-1])
    )


def _pep_mod_full_name(pep_seq, mods):
    return "-".join(
        letter + (
            " ({})".format(",".join(mods))
            if mods else
            ""
        )
        for letter, mods in zip(
            ["N-term"] + list(pep_seq) + ["C-term"],
            mods,
        )
    )


def _extract_mods(sequence):
    return tuple(
        tuple(mods)
        for _, mods in sequence
    )


def insert_ptm(cursor, query, seq, mod_state_id):
    mods = _extract_mods(seq)
    return _insert_or_update_row(
        cursor, "ptms", "ptm_id",
        {
            "mod_state_id": mod_state_id,
            "name": _pep_mod_name(query.pep_seq, mods),
            "full_name": _pep_mod_full_name(query.pep_seq, mods),
        },
        unique_on=["mod_state_id", "full_name"],
    )


def _ion_type_pos(name):
    name_match = regexes.RE_BY_ION_POS.match(name)
    if name_match:
        ion_pos = int(name_match.group(2))
        ion_type = "b" if name_match.group(1) in "abc" else "y"
    else:
        ion_type, ion_pos = None, None

    return ion_type, ion_pos


def insert_fragments(cursor, peaks, scan_ptm_id):
    gen = (
        (
            scan_ptm_id,
            peak_index,
            name,
            utils.rewrite_ion_name(name),
            mz,
            peak_hit.intensity,
            name == peak_hit.name,
        ) + _ion_type_pos(name)
        for peak_index, peak_hit in enumerate(peaks)
        if peak_hit.match_list
        for name, (mz, _) in peak_hit.match_list.items()
    )
    cursor.executemany(
        """
        INSERT OR IGNORE INTO fragments
        (
            scan_ptm_id,
            peak_id,
            name,
            display_name,
            mz,
            intensity,
            best,
            ion_type,
            ion_pos
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
        """,
        gen,
    )


def insert_camv_meta(cursor):
    cursor.executemany(
        """
        INSERT INTO camv_meta
        (
            key,
            val
        ) SELECT (?), (?)
        WHERE NOT EXISTS (SELECT 1 FROM camv_meta WHERE key=(?))
        """,
        [
            (i[0], i[1], i[0])
            for i in {
                "pycamverterVersion": version.__version__,
                "camvDataVersion": DATA_VERSION,
            }.items()
        ],
    )


def insert_path_data(cursor, search_path, raw_paths):
    cursor.executemany(
        """
        INSERT OR IGNORE INTO camv_meta
        (
            key,
            val
        ) VALUES (?, ?)
        """,
        [
            ("search_path", search_path),
        ] + [
            ("raw_path", raw_path)
            for raw_path in raw_paths
        ],
    )



"""
This module provides functionality for generating all possible combinations of
modifications on a peptide.
"""

from __future__ import absolute_import, division

import itertools


MAX_NUM_COMB = 10


def gen_possible_seq(pep_seq, var_mods):
    """
    Generates sequences

    Extra rules: Acetlyation cannot happen on terminal lysines, as trypsin does
    not cut at these sites.

    Parameters
    ----------
    pep_seq : str
    var_mods : list of tuple of (int, str, list of str)

    Returns
    -------
    generator of list of tuple of (str, list of str)
    """
    def _gen_mods(seq, mods):
        if not mods:
            yield seq
            return

        count, mod, letters = mods[0]

        # XXX: Allow multiple modifications on a peptide
        indices = [
            index
            for index, (letter, mods) in enumerate(seq)
            if letter in letters and not mods
        ]

        if len(indices) < count:
            raise Exception(
                (
                    "Too few sites for modification \"{} {} ({})\" "
                    "in peptide \"{}\""
                ).format(count, mod, "".join(letters), "".join(pep_seq[1:-1]))
            )

        for mod_indices in itertools.combinations(indices, count):
            tmp_seq = [
                (
                    letter,
                    (mods + (mod,)) if index in mod_indices else mods,
                )
                for index, (letter, mods) in enumerate(seq)
            ]

            for new_seq in _gen_mods(tmp_seq, mods[1:]):
                yield new_seq

    tmp_seq = ["N-term"] + list(pep_seq) + ["C-term"]

    # Sort var mods so that we assign more restricted sites first
    # (i.e. pY before pSTY)
    var_mods = sorted(var_mods, key=lambda x: len(x[2]))

    # Ideally we would use py>3.3 syntax: yield from, but we are also
    # supporting py2.7
    for seq in _gen_mods(
        seq=list(zip(tmp_seq, [() for _ in tmp_seq])),
        mods=var_mods,
    ):
        yield seq




import logging

from pycamv import version

LOGGER = logging.getLogger("pycamv.migrations")


def run_migrations(from_version, to_version, cursor):
    if from_version == "1.0.0":
        if to_version == "1.1.0":
            return migrate_1_0_0_to_1_1_0(cursor)

    raise NotImplementedError(
        "Unable to migrate CAMV data from {} to {}".format(
            from_version, to_version,
        )
    )


def migrate_1_0_0_to_1_1_0(cursor):
    LOGGER.info("Migrating data from 1.0.0 to 1.1.0")

    cursor.executescript("""
    ALTER TABLE proteins ADD COLUMN protein_uniprot text;
    ALTER TABLE proteins ADD COLUMN full_sequence   text;

    ALTER TABLE protein_peptide ADD COLUMN peptide_offset integer;

    ALTER TABLE protein_sets ADD COLUMN protein_set_uniprot text;

    ALTER TABLE peptides ADD COLUMN protein_set_offsets text;
    """)
    cursor.connection.commit()
    cursor.executemany(
        """
        UPDATE camv_meta
        SET val=(?)
        WHERE key=(?)
        """,
        [
            (i[1], i[0])
            for i in {
                "pycamverterVersion": version.__version__,
                "camvDataVersion": "1.1.0",
            }.items()
        ]
    )
    cursor.connection.commit()



PEPTIDE_LOSSES = []
INTERNAL_LOSSES = []
AA_LOSSES = {}
MOD_LOSSES = {
    # Losses from modified amino acids
    ("M", "Oxidation"): ["SOCH_4"],
    ("M", "Dioxidation"): ["SO_2CH_4"],
    ("S", "Phospho"): ["H_3PO_4"],
    ("T", "Phospho"): ["H_3PO_4"],
    ("Y", "Phospho"): ["HPO_3", "HPO_3-H_2O"],

    # Losses from unprotected N-/C-/internal fragment termini
    ("N-term", None): ["NH_3"],
    ("C-term", None): ["H_2O"],
    ("C=O", None): ["CO"],

    # Water losses from hydroxyl / carboxyl groups
    ("S", None): ["H_2O"],
    ("T", None): ["H_2O"],
    ("E", None): ["H_2O"],
    ("D", None): ["H_2O"],

    # Amine losses from amine groups
    ("R", None): ["NH_3"],
    ("K", None): ["NH_3"],
    ("Q", None): ["NH_3"],
    ("N", None): ["NH_3"],
}



"""Utility functions used in other modules."""

from __future__ import absolute_import, division

import copy
from collections import OrderedDict, Callable
import difflib
import itertools
import math

from . import regexes


def nCr(n, r):
    f = math.factorial
    return f(n) / f(r) / f(n - r)


class LenGen(object):
    def __init__(self, gen, len):
        self.gen = gen
        self.len = len

    def __call__(self):
        return itertools.islice(self.gen(), self.len)

    def __iter__(self):
        return self.gen

    def __len__(self):
        return self.len


class DefaultOrderedDict(OrderedDict):
    # Source: http://stackoverflow.com/a/6190500/562769
    def __init__(self, default_factory=None, *a, **kw):
        if (
            default_factory is not None and
            not isinstance(default_factory, Callable)
        ):
            raise TypeError("first argument must be callable")

        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()

        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,

        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(
            self.default_factory,
            self,
        )

    def __deepcopy__(self, memo):
        return type(self)(
            self.default_factory,
            copy.deepcopy(self.items()),
        )

    def __repr__(self):
        return "OrderedDefaultDict({}, {})".format(
            self.default_factory,
            OrderedDict.__repr__(self),
        )


class StrToBin(object):
    def __init__(self, f, encoding="utf-8"):
        self.f = f
        self.encoding = encoding

    def write(self, data):
        self.f.write(data.encode(self.encoding))


SUPERSCRIPT_UNICODE_START = ord(u"\u2070")
SUBSCRIPT_UNICODE_START = ord(u'\u2080')
SCRIPT_MAPPING = {
    str(i): i
    for i in range(10)
}
SCRIPT_MAPPING["+"] = 10
SCRIPT_MAPPING["-"] = 11
SCRIPT_MAPPING["("] = 12
SCRIPT_MAPPING[")"] = 13

try:
    unichr
except NameError:
    unichr = chr


def rewrite_ion_name(name):
    m = regexes.RE_B_Y_IONS.match(name)

    if m:
        name = "".join(m.group(1, 2))

    ret = ""
    sup, sub = False, False
    paren = 0

    for char in name:
        if char in "^_{}":
            if char == "^":
                sup, sub = True, False
            elif char == "_":
                sup, sub = False, True
            elif char == "}":
                sup, sub = False, False
                paren -= 1
            elif char == "{":
                paren += 1
            continue

        if sup:
            if char == "1":
                ret += u"\u00B9"
            elif char == "2":
                ret += u"\u00B2"
            elif char == "3":
                ret += u"\u00B3"
            else:
                ret += unichr(SUPERSCRIPT_UNICODE_START + SCRIPT_MAPPING[char])
        elif sub:
            ret += unichr(SUBSCRIPT_UNICODE_START + SCRIPT_MAPPING[char])
        else:
            ret += char

        if sup or sub:
            if not paren:
                sup, sub = False, False

    return ret


def fuzzy_find(needle, haystack):
    """
    Find the longest matching subsequence of needle within haystack.

    Returns the corresponding index from the beginning of needle.

    Parameters
    ----------
    needle : str
    haystack : str

    Returns
    -------
    int
    """
    s = difflib.SequenceMatcher(a=haystack, b=needle)
    best = s.find_longest_match(0, len(haystack), 0, len(needle))
    return best.a - len(needle) + best.size




from __future__ import absolute_import, division

import os

from pycamv.utils import nCr, fuzzy_find
from pycamv.fragment import ms_labels


class PeptideQuery:
    """
    Attributes
    ----------
    accessions : list of str
    prot_descs : list of str
    uniprot_accessions : list of str
    full_seqs : list of str
    pep_offsets : list of int
    query : int
    filename: str
    pep_score : float
    pep_exp_mz : float
    pep_exp_z : int
    pep_seq : str
    pep_var_mods : list of tuple of (int, str, tuple of str)
    pep_fixed_mods : list of tuple of (int, str, tuple of str)
    scan : int
    quant_scan : int or None
    num_comb : int
    rank_pos : set of tuple of int, str or None
    """
    def __init__(
        self,
        query,
        filename,
        pep_score,
        pep_exp_mz, pep_exp_z,
        pep_seq,
        pep_var_mods, pep_fixed_mods, scan,
        rank_pos=None,
        quant_scan=None,
        accessions=None,
        prot_descs=None,
        uniprot_accessions=None,
        full_seqs=None,
        pep_offsets=None,
    ):
        assert _check_mods(pep_var_mods)
        assert _check_mods(pep_fixed_mods)

        self.prot_descs = prot_descs or []
        self.accessions = accessions or []
        self.uniprot_accessions = uniprot_accessions or []
        self.full_seqs = full_seqs or []
        self.pep_offsets = pep_offsets or _get_offsets(pep_seq, full_seqs)

        self.query = query
        self.filename = filename
        self.pep_score = pep_score
        self.pep_exp_mz = pep_exp_mz
        self.pep_exp_z = pep_exp_z
        self.pep_seq = pep_seq.upper()
        self.pep_var_mods = pep_var_mods
        self.pep_fixed_mods = pep_fixed_mods
        self.scan = scan
        self.quant_scan = quant_scan
        self.num_comb = self._calc_num_comb()
        self.rank_pos = rank_pos

    def _unique_tuple(self):
        return (
            tuple(self.accessions),
            self.query,
            self.pep_seq,
            self.scan,
        )

    def __hash__(self):
        return hash(self._unique_tuple())

    def __eq__(self, other):
        if not isinstance(other, PeptideQuery):
            raise TypeError(other)
        return self._unique_tuple() == other._unique_tuple()

    @property
    def basename(self):
        return os.path.basename(self.filename)

    @property
    def prot_name(self):
        return " / ".join(sorted(self.prot_descs))

    @property
    def pep_mods(self):
        return self.pep_var_mods + self.pep_fixed_mods

    @property
    def get_label_mods(self):
        return [
            mod
            for _, mod, letters in self.pep_mods
            if mod in ms_labels.LABEL_NAMES and "N-term" in letters
        ]

    def _calc_num_comb(self):
        num_comb = 1

        for count, mod, letters in self.pep_var_mods:
            if mod == "Phospho" and letters == ["S", "T"]:
                letters = ["S", "T", "Y"]

            potential_mod_sites = sum(
                (list(self.pep_seq) + ["N-term", "C-term"]).count(i)
                for i in letters
            )

            # Subtract sites that will be taken up by another modification
            # (i.e. Oxidation and Dioxidation of M)
            for o_count, o_mod, o_letters in self.pep_var_mods:
                if o_letters == letters and o_mod != mod:
                    potential_mod_sites -= o_count

            num_comb *= nCr(potential_mod_sites, count)

        return num_comb


def _check_mods(mods):
    return all(
        isinstance(count, int) and
        isinstance(abbrev, str) and
        isinstance(letters, tuple) and
        all(isinstance(i, str) for i in letters)
        for count, abbrev, letters in mods
    )


def _get_offsets(pep_seq, full_seqs):
    def _get_rel_pos(seq):
        if not seq:
            return 0, False

        pep_pos = seq.find(pep_seq)
        exact = True

        if pep_pos < 0:
            pep_pos = fuzzy_find(pep_seq, seq)
            exact = False

        return pep_pos, exact

    return [_get_rel_pos(i)[0] for i in full_seqs]



"""
Provides functionality for interacting with MASCOT data.
"""

from __future__ import absolute_import, division

import logging
import os

from . import discoverer, mascot


LOGGER = logging.getLogger("pycamv.search")

BACKENDS = {
    ".xml": mascot.read_mascot_xml,
    ".msf": discoverer.read_discoverer_msf,
}


def read_search_file(path):
    """
    Parse a search input file.

    Parameters
    ----------
    path : str
        Path to search input file.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    """
    ext = os.path.splitext(path)[1]
    backend = BACKENDS.get(ext)

    if backend is None:
        raise Exception("Unable to open search file: {}".format(path))

    LOGGER.debug("Using {} backend for {}".format(backend.__name__, ext))

    return backend(path)




from __future__ import absolute_import, division

import logging
import os

from openpyxl import load_workbook

LOGGER = logging.getLogger("pycamv.scan_list")


def load_scan_xlsx(path):
    wb = load_workbook(path)
    ws = wb.active
    return [row[0].value for row in ws.iter_rows()]


def load_scan_csv(path):
    with open(path) as f:
        return [
            int(i.strip())
            for line in f
            for i in line.split(",")
        ]


BACKENDS = {
    ".csv": load_scan_csv,
    ".xlsx": load_scan_xlsx,
}


def load_scan_list(path):
    ext = os.path.splitext(path)[1]
    backend = BACKENDS.get(ext)

    if backend is None:
        raise Exception("Unable to open scan list: {}".format(path))

    LOGGER.debug("Using {} backend for {}".format(backend.__name__, ext))

    return backend(path)




import os
from setuptools import setup, find_packages

from pycamv import __version__


REQUIREMENTS_PATH = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__), "requirements.txt",
    )
)


setup(
    name="pycamverter",
    version=__version__,
    description=(
        "Utility for converting searched mass spec data into a format "
        "readable by CAMV"
    ),
    url="https://github.com/white-lab/pycamverter",
    author="Nader Morshed",
    author_email="morshed@mit.edu",
    license="BSD",
    packages=find_packages(exclude=["*.tests", "tests"]),
    install_requires=[
        "numpydoc>=0.7",
        "openpyxl>=2.5.0",
        "pymzml==0.7.7",
        "requests>=2.18.4",
        "scipy>=1.0.1",
    ],
    dependency_links=[
        "git+https://github.com/naderm/pymzML.git"
        "@master#egg=pymzml-0.7.7",
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        # "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Scientific/Engineering",
    ],
    test_suite="tests",
)



"""
Provides functionality for interacting with ProteomeDiscoverer data.
"""

from __future__ import absolute_import, division

from collections import Counter, defaultdict
from datetime import datetime
import logging
import os
import sqlite3
import re
import xml.etree.ElementTree as ET

from . import pep_query
from pycamv import regexes
from pycamv.fragment import masses


LOGGER = logging.getLogger("pycamv.discoverer")
RE_PSP = re.compile(r"(\w+)\((\d+)\): ([\d\.]+)")


def _parse_letters(letters):
    """
    Turns a string of residue letters (i.e. "STY") into a list.

    Includes special provisions for N- and C-term modifications.

    Parameters
    ----------
    letters : str

    Returns
    -------
    tuple of str
    """
    if letters in ["N-term", "C-term"]:
        return (letters,)

    return tuple(letters)


def _find_mod(abbrev, letter, pot_mods):
    for pot_mod in pot_mods:
        if letter not in pot_mod[1]:
            continue

        if pot_mod[0] in ['TMT6plex', 'TMT10plex']:
            mods = ['TMT6plex', 'TMT10plex']
        else:
            mods = [pot_mod[0]]

        if abbrev in mods:
            return pot_mod

    return None


def _count_mods(mod_list):
    return [
        (count, abbrev, _parse_letters(letters))
        for (abbrev, letters), count in Counter(mod_list).items()
    ]


def _get_fixed_var_mods(conn, pd_version):
    if pd_version[:2] in [(1, 4)]:
        query = conn.cursor().execute(
            """
            SELECT
            ProcessingNodeParameters.ParameterName,
            ProcessingNodeParameters.ParameterValue

            FROM
            ProcessingNodeParameters
            """,
        )

        fixed_mods = []
        var_mods = []

        for name, val in query:
            if name.startswith("StaticMod_"):
                if val.strip():
                    fixed_mods.append(val.strip())
            elif name.startswith("DynMod_"):
                if val.strip():
                    var_mods.append(val.strip())

        return fixed_mods, var_mods
    elif pd_version[:2] in [(2, 1), (2, 2)]:
        query = conn.cursor().execute(
            """
            SELECT
            Workflows.WorkflowXML

            FROM
            Workflows
            """
        )

        fixed_mods = []
        var_mods = []

        for xml, in query:
            root = ET.fromstring(xml)
            params = root.findall(
                "WorkflowTree/WorkflowNode/"
                "ProcessingNodeParameters/ProcessingNodeParameter"
            )
            for param in params:
                if not param.get("IsValueSet", False) or not param.text:
                    continue

                if param.get("Name", "").startswith("DynMod_"):
                    if param.text.strip():
                        var_mods.append(param.text.strip())
                elif param.get("Name", "").startswith("StaticMod_"):
                    if param.text.strip():
                        fixed_mods.append(param.text.strip())

        return fixed_mods, var_mods
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_aa_mods(conn, pep_id, pd_version):
    if pd_version[:2] in [(1, 4), (2, 1)]:
        return conn.cursor().execute(
            """
            SELECT
            Peptides.SearchEngineRank,
            AminoAcidModifications.ModificationName,
            PeptidesAminoAcidModifications.Position

            FROM Peptides

            JOIN PeptidesAminoAcidModifications
            ON PeptidesAminoAcidModifications.PeptideID=Peptides.PeptideID

            JOIN AminoAcidModifications
            ON PeptidesAminoAcidModifications.AminoAcidModificationID=
            AminoAcidModifications.AminoAcidModificationID

            WHERE
            Peptides.PeptideID=:pepID
            """,
            {
                "pepID": pep_id,
            }
        )
    elif pd_version[:2] in [(2, 2)]:
        return conn.cursor().execute(
            """
            SELECT
            TargetPsms.SearchEngineRank,
            FoundModifications.Name,
            TargetPsmsFoundModifications.Position

            FROM TargetPsms

            JOIN TargetPsmsFoundModifications
            ON
            TargetPsmsFoundModifications.TargetPsmsPeptideID=TargetPsms.PeptideID

            JOIN FoundModifications
            ON
            TargetPsmsFoundModifications.FoundModificationsModificationID=
            FoundModifications.ModificationID

            WHERE
            TargetPsms.PeptideID=:pepID AND
            FoundModifications.PositionType NOT IN (1, 2)
            """,
            {
                "pepID": pep_id,
            }
        )
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_term_mods(conn, pep_id, pd_version):
    if pd_version[:2] in [(1, 4), (2, 1)]:
        return conn.cursor().execute(
            """
            SELECT
            AminoAcidModifications.ModificationName,
            AminoAcidModifications.PositionType

            FROM PeptidesTerminalModifications

            JOIN AminoAcidModifications
            ON PeptidesTerminalModifications.TerminalModificationID=
            AminoAcidModifications.AminoAcidModificationID

            WHERE
            PeptidesTerminalModifications.PeptideID=:pepID
            """,
            {
                "pepID": pep_id,
            }
        )
    elif pd_version[:2] in [(2, 2)]:
        return conn.cursor().execute(
            """
            SELECT
            FoundModifications.Name,
            FoundModifications.PositionType

            FROM TargetPsmsFoundModifications

            JOIN FoundModifications
            ON
            TargetPsmsFoundModifications.FoundModificationsModificationID=
            FoundModifications.ModificationID

            WHERE
            TargetPsmsFoundModifications.TargetPsmsPeptideID=:pepID AND
            FoundModifications.PositionType IN (1, 2)
            """,
            {
                "pepID": pep_id,
            }
        )
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_proteins(conn, peptide_id, pd_version):
    if pd_version[:2] in [(1, 4), (2, 1)]:
        return conn.cursor().execute(
            """
            SELECT
            ProteinAnnotations.Description,
            Proteins.Sequence

            FROM PeptidesProteins

            JOIN ProteinAnnotations
            ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID

            JOIN Proteins
            ON Proteins.ProteinID=PeptidesProteins.ProteinID

            WHERE
            PeptidesProteins.PeptideID=:pepID
            """,
            {
                "pepID": peptide_id,
            }
        )
    elif pd_version[:2] in [(2, 2)]:
        return conn.cursor().execute(
            """
            SELECT
            TargetProteins.FastaTitleLines,
            TargetProteins.Sequence

            FROM TargetProteinsTargetPsms

            JOIN TargetProteins
            ON TargetProteins.UniqueSequenceID=
            TargetProteinsTargetPsms.TargetProteinsUniqueSequenceID

            WHERE
            TargetProteinsTargetPsms.TargetPsmsPeptideID=:pepID
            """,
            {
                "pepID": peptide_id,
            }
        )
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_quant_scan(conn, scan, pd_version):
    if pd_version[:2] in [(1, 4), (2, 1)]:
        query = conn.cursor().execute(
            """
            SELECT
            quant_scans.FirstScan

            FROM ReporterIonQuanResultsSearchSpectra mapping

            JOIN SpectrumHeaders quant_scans
            ON mapping.SpectrumID=quant_scans.SpectrumID

            JOIN SpectrumHeaders search_scans
            ON search_scans.SpectrumID=mapping.SearchSpectrumID

            WHERE search_scans.FirstScan=:scanId
            """,
            {
                "scanId": scan,
            }
        )
        for quant_scan, in query:
            return quant_scan
    elif pd_version[:2] in [(2, 2)]:
        query = conn.cursor().execute(
            """
            SELECT
            QuanSpectrumInfo.FirstScan

            FROM MSnSpectrumInfo

            JOIN QuanSpectrumInfoMSnSpectrumInfo
            ON MSnSpectrumInfo.SpectrumID=
            QuanSpectrumInfoMSnSpectrumInfo.MSnSpectrumInfoSpectrumID

            JOIN QuanSpectrumInfo
            ON QuanSpectrumInfo.SpectrumID=
            QuanSpectrumInfoMSnSpectrumInfo.QuanSpectrumInfoSpectrumID

            WHERE MSnSpectrumInfo.FirstScan=:scanId
            """,
            {
                "scanId": scan,
            }
        )
        for quant_scan, in query:
            return quant_scan
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_pep_info(conn, pd_version):
    if pd_version[:2] in [(1, 4)]:
        return conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            SpectrumHeaders.Charge,
            FileInfos.FileName

            FROM Peptides

            JOIN PeptideScores
            ON PeptideScores.PeptideID=Peptides.PeptideID

            JOIN PeptidesProteins
            ON PeptidesProteins.PeptideID=Peptides.PeptideID

            JOIN ProteinAnnotations
            ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID

            JOIN SpectrumHeaders
            ON SpectrumHeaders.SpectrumID=Peptides.SpectrumID

            JOIN FileInfos
            ON FileInfos.FileID=MassPeaks.FileID

            JOIN Masspeaks
            ON Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID

            WHERE
            Peptides.SearchEngineRank=1
            """
        )
    elif pd_version[:2] in [(2, 1)]:
        return conn.cursor().execute(
            """
            SELECT
            Peptides.PeptideID,
            Peptides.Sequence,
            PeptideScores.ScoreValue,
            SpectrumHeaders.FirstScan,
            Masspeaks.Mass,
            Masspeaks.Charge,
            SpectrumHeaders.Charge,
            WorkflowInputFiles.FileName

            FROM Peptides

            JOIN PeptideScores
            ON PeptideScores.PeptideID=Peptides.PeptideID

            JOIN PeptidesProteins
            ON PeptidesProteins.PeptideID=Peptides.PeptideID

            JOIN ProteinAnnotations
            ON ProteinAnnotations.ProteinID=PeptidesProteins.ProteinID

            JOIN SpectrumHeaders
            ON SpectrumHeaders.SpectrumID=Peptides.SpectrumID

            JOIN WorkflowInputFiles
            ON WorkflowInputFiles.FileID=MassPeaks.FileID

            JOIN Masspeaks
            ON Masspeaks.MassPeakID=SpectrumHeaders.MassPeakID

            WHERE
            Peptides.SearchEngineRank=1
            """
        )
    elif pd_version[:2] in [(2, 2)]:
        return conn.cursor().execute(
            """
            SELECT
            TargetPsms.PeptideID,
            TargetPsms.Sequence,
            TargetPsms.IonsScore,
            TargetPsms.FirstScan,
            TargetPsms.Mass,
            TargetPsms.OriginalPrecursorCharge,
            TargetPsms.Charge,
            WorkflowInputFiles.FileName

            FROM TargetPsms

            JOIN WorkflowInputFiles
            ON WorkflowInputFiles.FileID=TargetPsms.SpectrumFileID

            WHERE
            TargetPsms.SearchEngineRank=1
            """
        )
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_pep_mods(conn, pep_id, pep_seq, var_mods, fixed_mods, pd_version):
    aa_mods = _get_aa_mods(conn, pep_id, pd_version)

    pep_var_mods = []
    pep_fixed_mods = []
    rank_pos = defaultdict(set)

    for rank, abbrev, pos in sorted(
        aa_mods,
        key=lambda x: 0 if x[1].startswith("Mapping") else 1,
    ):
        if pd_version[:2] in [(2, 2)]:
            pos -= 1

        letter = pep_seq[pos]
        rank_pos[rank].add((pos, abbrev))

        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        # Weird ProteomeDiscoverer setting for proteins with unknown amino
        # acids. We can remap them to their correct amino acid:
        if (
            letter == "X" and
            (letter, abbrev) in masses.MODIFICATIONS and
            abbrev.startswith("Mapping")
        ):
            pep_seq = pep_seq[:pos] + abbrev[-1] + pep_seq[pos + 1:]
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    term_mods = _get_term_mods(conn, pep_id, pd_version)

    for abbrev, pos_type in term_mods:
        letter = "N-term" if pos_type == 1 else "C-term"
        mod = _find_mod(abbrev, letter, var_mods)

        if mod:
            pep_var_mods.append(mod)
            continue

        mod = _find_mod(abbrev, letter, fixed_mods)

        if mod:
            pep_fixed_mods.append(mod)
            continue

        raise Exception(
            "Unexpected modification: {} {}".format(letter, abbrev)
        )

    pep_var_mods = _count_mods(pep_var_mods)
    pep_fixed_mods = _count_mods(pep_fixed_mods)

    return pep_seq, pep_var_mods, pep_fixed_mods, rank_pos


def _get_prot_info(conn, peptide_id, pd_version):
    proteins = _get_proteins(conn, peptide_id, pd_version)

    (
        accessions,
        descriptions,
        uniprot_accessions,
        full_seqs,
    ) = [], [], [], []

    for full_prot_desc, sequence in proteins:
        for line in full_prot_desc.split('\n'):
            uniprot_accession, accession, prot_desc = \
                regexes.RE_DISCOVERER_DESCRIPTION.match(line).group(1, 2, 3)

            if not accession:
                raise Exception(
                    "Unable to find accession ID for {}".format(full_prot_desc)
                )

            accessions.append(accession)
            descriptions.append(prot_desc)
            uniprot_accessions.append(uniprot_accession)
            full_seqs.append(sequence)

    return (
        accessions,
        descriptions,
        uniprot_accessions,
        full_seqs,
    )


def _get_peptide_queries(conn, fixed_mods, var_mods, pd_version):
    out = []
    # index = 0
    # scan_used = {}
    fixed_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in fixed_mods
    ]
    var_mods = [
        regexes.RE_DYN_MODS.match(i).group(3, 4)
        for i in var_mods
    ]

    query = _get_pep_info(conn, pd_version)

    psp_vals = _get_phosphors_psp_vals(conn.cursor(), pd_version)
    changed_peptides = 0
    ambiguous_peptides = 0

    for (
        pep_id, pep_seq, pep_score,
        scan, exp_mz, mass_z, exp_z, filename,
    ) in query:
        if exp_z < 1:
            LOGGER.warning(
                "Charge: {}, {} for {} (Scan {})"
                .format(exp_z, mass_z, pep_seq, scan)
            )

        pep_seq, pep_var_mods, pep_fixed_mods, rank_pos = _get_pep_mods(
            conn,
            pep_id,
            pep_seq,
            var_mods,
            fixed_mods,
            pd_version,
        )

        quant_scan = _get_quant_scan(conn, scan, pd_version)

        (
            accessions,
            descriptions,
            uniprot_accessions,
            full_seqs,
        ) = _get_prot_info(conn, pep_id, pd_version)

        if pep_id in psp_vals:
            rank_pos, reassigned, ambiguous = _reassign_rank(
                pep_fixed_mods + pep_var_mods,
                rank_pos,
                psp_vals[pep_id],
            )

            if reassigned:
                changed_peptides += 1

            if ambiguous:
                ambiguous_peptides += 1

        out.append(
            pep_query.PeptideQuery(
                accessions=accessions,
                prot_descs=descriptions,
                uniprot_accessions=uniprot_accessions,
                full_seqs=full_seqs,
                query=pep_id,
                filename=filename,
                pep_score=pep_score,
                pep_exp_mz=exp_mz,
                pep_exp_z=exp_z,
                pep_seq=pep_seq,
                pep_var_mods=pep_var_mods,
                pep_fixed_mods=pep_fixed_mods,
                scan=scan,
                rank_pos=rank_pos,
                quant_scan=quant_scan,
            )
        )

    out = sorted(out, key=lambda x: x.scan)

    LOGGER.info(
        "Reassigned {} top modification assignments using phosphoRS"
        .format(changed_peptides)
    )

    return out


def _is_pmod(mod):
    pos, mod_type = mod
    return (
        mod_type in ["Phospho"]
    )


def _sort_mods(mods):
    return tuple(
        sorted(
            mods,
            key=lambda x: (x[0], x[1]),
        )
    )


def _reassign_rank(mods, rank_pos, psp_val):
    reassigned = False
    ambiguous = False

    # phophoRS example format: "T(4): 99.6; S(6): 0.4; S(10): 0.0"
    # Error messages include: "Too many isoforms"
    psp_val = [
        RE_PSP.match(i.strip())
        for i in psp_val.split(";")
    ]
    psp_val = [
        i.groups()
        for i in psp_val
        if i
    ]
    psp_val = [
        (i[0], int(i[1]), float(i[2]))
        for i in psp_val
    ]

    if 1 not in rank_pos:
        return rank_pos, False, False

    o_mods = [i for i in rank_pos[1] if not _is_pmod(i)]
    p_mods = [i for i in rank_pos[1] if _is_pmod(i)]
    psp_val_f = [i for i in psp_val if i[2] > 50]

    if len(p_mods) != len(psp_val_f):
        LOGGER.debug(
            "Not enough info to assign phophosite: {}".format(psp_val)
        )
        ambiguous = True
        rank_pos = None
    elif set(i[0] + 1 for i in p_mods) != set(i[1] for i in psp_val_f):
        p_mods = [
            (i[1] - 1, "Phospho")
            for i in psp_val_f
        ]
        reassigned = True

        rank_pos = _sort_mods(o_mods + p_mods)

    return rank_pos, reassigned, ambiguous


def _get_phosphors_psp_vals(cursor, pd_version):
    if pd_version in [(1, 4), (2, 1)]:
        fields = cursor.execute(
            """
            SELECT
            CustomDataFields.FieldID,
            CustomDataFields.DisplayName
            FROM CustomDataFields
            """,
        )
        field_ids = [
            field_id
            for field_id, name in fields
            if name in ["phosphoRS Site Probabilities"]
        ]

        if not field_ids:
            return {}

        psp_vals = cursor.execute(
            """
            SELECT
            CustomDataPeptides.PeptideID,
            CustomDataPeptides.FieldValue
            FROM CustomDataPeptides
            WHERE CustomDataPeptides.FieldID IN ({})
            """.format(
                ", ".join("?" * len(field_ids))
            ),
            field_ids,
        )
        return {
            pep_id: psp_val
            for pep_id, psp_val in psp_vals
        }
    elif pd_version[:2] in [(2, 2)]:
        return {}
    else:
        raise Exception(
            "Unknown Proteome Discoverer Version: {}".format(pd_version)
        )


def _get_pd_version(conn):
    query = conn.cursor().execute(
        """
        SELECT
        SoftwareVersion
        FROM SchemaInfo
        WHERE Kind=='Result'
        """,
    )
    return tuple([int(i) for i in next(query)[0].split('.')])


def read_discoverer_msf(msf_path):
    """
    Parse a ProteomeDiscoverer MSF file.

    Parameters
    ----------
    msf_path : str
        Path to MSF file.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.pep_query.PeptideQuery>`
    """
    LOGGER.info(
        "Loading ProteomeDiscoverer peptides from \"{}\"".format(
            os.path.basename(msf_path),
        )
    )
    start = datetime.now()

    with sqlite3.connect(msf_path) as conn:
        pd_version = _get_pd_version(conn)
        fixed_mods, var_mods = _get_fixed_var_mods(
            conn,
            pd_version=pd_version,
        )
        out = _get_peptide_queries(
            conn,
            fixed_mods,
            var_mods,
            pd_version=pd_version,
        )

    LOGGER.info(
        "Loaded {} peptides in {} hr:min:sec"
        .format(
            len(out),
            str(datetime.now() - start).split('.')[0],
        )
    )

    return fixed_mods, var_mods, out
